<?php
// $Id: adserve.php,v 1.2.2.15.2.6 2007/04/30 13:17:53 jeremy Exp $

adserve();

/**
 * The main adserve logic.
 */
function adserve() {
  adserve_variable('ad_dir', getcwd());

  // Move process to top level Drupal directory so we can perform a bootstrap.
  for ($i = 1; $i <= 20; $i++) {
    if (file_exists('./includes/bootstrap.inc')) {
      break;
    }
    chdir('..');
  }
  adserve_variable('root_dir', getcwd());
  require_once adserve_variable('root_dir') .'/includes/bootstrap.inc';

  adserve_debug();

  switch (adserve_variable('cache')) {
    case 'none':
    {
      if (adserve_variable('debug')) {
        echo "Using no cache.<br />\n";
      }
  
      adserve_bootstrap();
  
      if (adserve_variable('nids')) {
        $nids = explode(',', adserve_variable('nids'));
  
        // Build query to filter on just selected nids.
        $filter = '';
        foreach ($nids as $nid) {
          if ($filter) {
            $filter .= " OR aid = $nid";
          }
          else {
            $filter = " AND (aid = $nid";
          }
        }
        $filter .= ')';
  
        // Randomly determine which ad to display from the specified nids.
        $result = db_query("SELECT aid, adtype, redirect, activated, maxviews, maxclicks FROM {ads} a WHERE adstatus = 'active' $filter ORDER BY RAND() LIMIT %d", adserve_variable('quantity'));
      }
      else {
        // Randomly determine which ad to display from the specified group.
        $result = db_query("SELECT a.aid, a.adtype, a.redirect, a.activated, a.maxviews, a.maxclicks FROM {ads} a INNER JOIN {ad_groups} g ON a.gid = g.gid WHERE a.adstatus = 'active' AND g.name = '%s' ORDER BY RAND() LIMIT %d", adserve_variable('group'), adserve_variable('quantity'));
      }
      
      $output = '';
      $ads = 0;
      // Include appropriate module for displaying selected ad.
      while ($ad = db_fetch_object($result)) {
        $ads++;
        $module = db_result(db_query("SELECT filename FROM {system} WHERE name = '%s'", "ad_$ad->adtype"));
        if (adserve_variable('debug')) {
            echo "ad: <pre>";
            print_r($ad);
            echo "</pre>";
            echo "Loading module '$module'.<br />\n";
        }
        include_once "$module";
      
        if ($output) {
          // TODO: Make this configurable or themeable.
          $output .= '<br />';
        }
        $output .= module_invoke("ad_$ad->adtype", 'display_ad', $ad, adserve_variable('hostid'));
  
        // Update view statistics.
        db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = 'view' AND date = %d AND hostid = '%s'", $ad->aid, date('YmdH'), adserve_variable('hostid'));
        // If column doesn't already exist, we need to add it.
        if (!db_affected_rows()) {
          db_query("INSERT INTO {ad_statistics} (aid, date, action, hostid, count) VALUES(%d, %d, 'view', '%s', 1)", $ad->aid, date('YmdH'), adserve_variable('hostid'));
          // If another process already added this row our INSERT will fail, if 
          // so we still need to increment it so we don't loose a view.
          if (!db_affected_rows()) {
            db_query("UPDATE {ad_statistics} SET count = count + 1 WHERE aid = %d AND action = 'view' AND date = %d AND hostid = '%s'", $ad->aid, date('YmdH'), adserve_variable('hostid'));
          }
        }
  
        // See if we need to perform additional queries.
        if ($ad->maxviews) {
          $views = (int)db_result(db_query("SELECT SUM(count) FROM {ad_statistics} WHERE aid = %d AND action = 'view' AND date >= %d", $ad->aid, date('YmdH', $ad->activated)));
          if ($views >= $ad->maxviews) {
            db_query("UPDATE {ads} SET adstatus = 'expired', autoexpire = 0, autoexpired = %d, expired = %d WHERE aid = %d", time(), time(), $ad->aid);
            ad_statistics_increment($ad->aid, 'autoexpired');
            ad_statistics_increment($ad->aid, 'expired');
          }
        }
        if ($ad->maxclicks) {
          $clicks = (int)db_result(db_query("SELECT SUM(count) FROM {ad_statistics} WHERE aid = %d AND action = 'click' AND date >= %d", $ad->aid, date('YmdH', $ad->activated)));
          if ($clicks >= $ad->maxclicks) {
            db_query("UPDATE {ads} SET adstatus = 'expired', autoexpire = 0, autoexpired = %d, expired = %d WHERE aid = %d", time(), time(), $ad->aid);
            ad_statistics_increment($ad->aid, 'autoexpired');
            ad_statistics_increment($ad->aid, 'expired');
          }
        }
  
      }
      if (!$ads) {
        $output = "No ad found.";
      }
      if (adserve_variable('debug')) {
        echo "Ads displayed: $ads<br />";
      }
      break;
    }
  
    case 'file':
    {
      if (adserve_variable('debug')) {
        echo "Using file cache.<br />\n";
      }
  
      // We'll loop through all possible cache files until we obtain an 
      // exclusive lock.
      $lock = FALSE;
      for ($i = 1; $i <= adserve_variable('files'); $i++) {
        // Prefix the filename with a '.' to hide it on Unix systems.
        $cache_file = adserve_variable('path') .'/.'. $i .'.ad.cache';
  
        if (adserve_variable('debug')) {
          echo "Trying cache_file '$cache_file'.<br />\n";
        }
  
        if (!$fd = @fopen($cache_file, 'r+')) {
          if (adserve_variable('debug')) {
            echo "Failed to open cache_file '$cache_file'.<br />\n";
          }
          // We failed to open the cache file, try the next one.
          continue;
        }
        if ($i < adserve_variable('files')) {
          // This isn't the last available cache file so we'll use a 
          // non-blocking lock for best performance.  If we fail to lock this 
          // cache file, we'll quickly move on to the next until we find an 
          // available one.
          if (!flock($fd, LOCK_EX|LOCK_NB)) {
            if (adserve_variable('debug')) {
              echo "Failed to obtain non-blocking lock.<br />\n";
            }
            // We failed to obtain an exclusive lock, close the file and try the
            // next one.
            @fclose($fd);
            continue;
          }
          if (adserve_variable('debug')) {
            echo "Obtained lock.<br />\n";
          }
          $lock = TRUE;
          break;
        }
        else {
          // This is the last available cache file, we'll use a blocking lock 
          // as we have to wait until we have exclusive write permissions.
          if (!flock($fd, LOCK_EX)) {
            if (adserve_variable('debug')) {
              echo "Failed to obtain blocking lock.<br />\n";
            }
            // A blocking exclusive lock shouldn't ever fail, so something has
            // gone very wrong.  Perhaps the file was deleted out from under us?
            @fclose($fd);
            continue;
          }
          if (adserve_variable('debug')) {
            echo "Obtained lock.<br />\n";
          }
          $lock = TRUE;
          break;
        }
      }
  
      if ($lock) {
        // Read entire cache file into memory.
        $cache = unserialize(fread($fd, filesize($cache_file)));
        $group = adserve_variable('group');
  
        /**
         * The cache structure looks like this:
         * $cache['ad'][$group][]['display'] = $ad
         * $cache['ad'][$group][]['aid'] = $aid
         * $cache['ad'][$group][][$action][$timestamp] = $counter
         * $cache['last_sync'] = $timestamp
         */
  
        $output = '';
        $ids = array();
        for ($i = 0; $i < adserve_variable('quantity'); $i++) {
          if (adserve_variable('nids')) {
            $nids = explode(',', adserve_variable('nids'));
            $cache_size = sizeof($nids);
          }
          else {
            $cache_size = sizeof($cache['ad'][$group]);
          }
  
          if (adserve_variable('debug')) {
            echo "$i: cache_size($cache_size)<br />";
          }
  
          // We can't display more than $cache_size unique ads.
          if ($i >= $cache_size) 
            break;
          $id = FALSE;
  
          while ($id === FALSE || in_array($id, $ids)) {
          // Randomly select a unique banner to display.  We subtract 1 as 
          // arrays start at 0.
            $id = $cache_size > 1 ? mt_rand(0, $cache_size - 1) : 0;
            if (adserve_variable('debug')) {
              echo "Id: $id<br />";
            }
          }
  
          if ($nids) {
            $group = FALSE;
            $loop = 0;
            while (!$group) {
              if (adserve_variable('debug')) {
                echo "$loop: random id: $id<br />";
              }
              $group = $cache['ad'][$nids[$id]]['group'];
              if (!$group) {
                // We were passed an invalid nid, walk through all available to
                // find one that is valid.  The ad that comes after a 
                // nonexistent ad will end up being favored by average, but so 
                // it goes...
                if ($id == $cache_size) {
                  $id = 0;
                }
                else {
                  $id++;
                }
              // If all passed in nid were invalid, prevent an infinite loop.
              if ($loop++ == $cache_size)
                break;
              }
            }
            $ad_id = $cache['ad'][$nids[$id]]['id'];
            $ids[] = $id;
            $id = $ad_id;
          }
          else {
            $ids[] = $id;
          }

          $ad = $cache['ad'][$group][$id];
          $last_sync = $cache['last_sync'];

          if ($output) {
            $output .= '<br />';
          }
          $output .= $ad['display'];
  
          if (adserve_variable('debug')) {
            echo "Cache size: $cache_size<br />\n";
            echo "Random ID: $id<br />\n";
            $aid = $cache['ad'][$group][$id]['aid'];
            echo "Ad ID: $aid<br />\n";
            echo "Last sync: $last_sync<br />\n";
          }

          // Increment view counter.
          $timestamp = date('YmdH');
          if (isset($cache['ad'][$group][$id]['view'])) {
            $cache['ad'][$group][$id]['view'][$timestamp]++;
          }
          else {
            $cache['ad'][$group][$id]['view'][$timestamp] = 1;
          }
        }

        // Write updated cache back to file and release the lock.
        $cache = serialize($cache);
        rewind($fd);
        fwrite($fd, $cache, strlen($cache));
        flock($fd, LOCK_UN);
        fclose($fd);

        // Every 60 seconds we flush the cache files to the database.
        if ($last_sync < time() - 60) {
          adserve_rebuild_cache_files();
        }
      }
      else {
        $output = 'Failed to load image from cache file.';
        adserve_rebuild_cache_files();
      }
      if ($output == '') {
        $output = "No ad found.";
      }
      break;
    }

    default:
    {
      /**
       * It's possible to define your own caching mechanism.  You simply have
       * to put all necessary logic in a file named "ad_cache_TYPE.module" where
       * TYPE is your type of cache.  For example, if you had a "memory" cache
       * you would create a file called "ad_cache_memory.module".  It must have
       * a function called ad_cache_TYPE(), so with our "memory" example above
       * the function would be named "ad_cache_memory()".  That function in
       * your cache file will then be called, and will have access to all the
       * global variables defined at the top of this file. (ie $group, $hostid,
       * etc...)  Your function must return the ad exactly as it is to be
       * displayed.
       */
      $type = "ad_cache_$cache";
      $module = adserve_variable('ad_dir') ."/$type.module";
      if (file_exists($module)) {
        require_once $module;
        if (function_exists($type)) {
          $output = $type();
        }
        else {
          $output = "Required function $type() does not exist.";
        }
      }
      else {
        $output = "Failed to include file $module.";
      }
    }
  }
  header("Content-Type: application/x-javascript; charset=utf-8");
  print "document.write('". $output ."');";
  exit(0);
}


/**
 * Retrieve variables from $_GET array.
 */
function adserve_variable($variable, $value = NULL) {
  static $variables = NULL;

  if (isset($value)) {
    $variables->$variable = $value;
  }

  if (!isset($variables->loaded)) {
    // 'debug' is an integer.
    $variables->debug = isset($_GET['debug']) ? (int)$_GET['debug'] : 0;
    // group names are comprised of only alphanumerics.
    $variables->group = isset($_GET['g']) ? preg_replace('/[^0-9a-zA-Z]/', '', $_GET['g']) : 'default';
    // cache types are comprised of only letters.
    $variables->cache = isset($_GET['c']) ? preg_replace('/[^a-zA-Z]/', '', $_GET['c']) : 'none';
    // nids is an integer or a ",".
    $variables->nids = isset($_GET['n']) ? preg_replace('/[^0-9,]/', '', $_GET['n']) : '';
    // hostid is an md5() which is comprised of numbers and letters a-f.
    $variables->hostid = isset($_GET['k']) ? preg_replace('/[^0-9a-f]/', '', $_GET['k']) : '';

    // paths are comprised of alphanumerics, underscores, dashes, periods and 
    // slashes.
    $variables->path = isset($_GET['p']) ? preg_replace('/[^_\-\.\/\\0-9a-zA-Z]/', '', $_GET['p']) : '';
    // files is an integer.
    $variables->files = isset($_GET['f']) ? (int)$_GET['f'] : 0;
    // quantity is an integer.
    $variables->quantity = isset($_GET['q']) ? (int)$_GET['q'] : 0;

    // Set defaults.
    if ($variables->files > 15) {
      echo "Invalid value 'f=$variables->files', exiting.";
      exit();
    }
    $variables->quantity = $variables->quantity ? $variables->quantity : 1;

    if ($variables->debug) {
      echo "Group: '$variables->group'<br />\n";
      echo "Cache: '$variables->cache'<br />\n";
      echo "Quantity: '$variables->quantity'<br />\n";
      echo "Host: '$variables->hostid'<br />\n";
      echo "Path: '$variables->path'<br />\n";
      echo "Nids: '$variables->nids'<br />\n";
      echo "Files: '$variables->files'<br />\n";
      if ($variables->debug == 1) exit;
    }
    $variables->loaded = TRUE;
  }

  return $variables->$variable;
}

/**
 * Display additional debug information.
 */
function adserve_debug() {
  if (adserve_variable('debug')) {
    echo "<br />\nAd module installed in '". adserve_variable('ad_dir') ."'.<br />\n";
    echo "Root drupal directory '". adserve_variable('root_dir') ."'.<br />\n<br />\n";

    $ad_dir = adserve_variable('ad_dir');
    $files = array("$ad_dir/adserve.php", "$ad_dir/ad.module");
    if (adserve_variable('debug') > 2) {
      $files = array_merge($files, array("$ad_dir/ad.install", "$ad_dir/ad_views.inc"));
    }
    if (adserve_variable('debug') > 3) {
      $files = array_merge($files, array("$ad_dir/image/ad_image.module", "$ad_dir/image/ad_image.install", "$ad_dir/image/ad_image_views.inc", "$ad_dir/text/ad_text.module", "$ad_dir/text/ad_text.install", "$ad_dir/text/ad_text_views.inc", "$ad_dir/embed/ad_embed.module", "$ad_dir/report/ad_report.module", "$ad_dir/notify/ad_notify.module", "$ad_dir/notify/ad_notify.install"));
    }
    foreach ($files as $file) {
      if (!file_exists($file)) {
        echo "Error: '$file' does not exist!<br />\n";
      }
      else if (!is_readable($file)) {
        echo "Error: '$file' is not readable!<br />\n";
      }
      else {
        $fd = fopen($file, 'r');
        while (!feof($fd)) {
          $line = fgets($fd);
          if (substr($line, 0, 5) == "<?php") {
            continue;
          }
          else {
            echo "$file: $line<br />";
            break;
          }
        }
      }
    }
    echo "<br />\n";
  }
}

/**
 * Bootstrap drupal, then run _ad_build_cache_files() from ad.module which will 
 * rebuild all cache files.
 */
function adserve_rebuild_cache_files() {
  adserve_bootstrap();
  _ad_build_cache_files();
}

/**
 * Include the necessary files and call the Drupal bootstrap.
 */
function adserve_bootstrap($bootstrap = DRUPAL_BOOTSTRAP_FULL) {
  if (adserve_variable('debug')) {
    echo "Drupal bootstrap '". $bootstrap ."'.<br />\n";
  }

  drupal_bootstrap($bootstrap);
}

?>
